# -----------------------------------------------------------------------------
# Generate resonance and peak assignments for proteins using sequence and
# peak lists for several spectra.
#
import Tkinter
import types

import experiments
import pyutil
import sequence
import shiftstats
import sparky
import spingraph
import sputil
import tkutil

# -----------------------------------------------------------------------------
# Select experiments and display possible assignment extensions on spin graph.
#
class assigner_dialog(tkutil.Dialog, tkutil.Stoppable):

  def __init__(self):

    self.assignments = None
    self.spin_graph = None
    self.current_resonance = None
    self.setup = assigner_setup_dialog(self)
    self.shown_group_atoms = None
    
    tkutil.Dialog.__init__(self, 'Assignment Graph')

    lb = self.peak_list_box(self.top)
    lb.pack(side = 'top', fill = 'both', expand = 1)
    
    progress_label = Tkinter.Label(self.top, anchor = 'nw')
    progress_label.pack(side = 'top', anchor = 'w')

    br = tkutil.button_row(self.top,
			   ('Assign', self.assign_cb),
                           ('Unassign', self.unassign_cb),
                           ('Update', self.update_cb),
			   ('Setup', self.setup_cb),
                           ('Stop', self.stop_cb),
                           ('Close', self.close_cb),
                           ('Help', sputil.help_cb('AssignGraph')),
			   )
    br.frame.pack(side = 'top', anchor = 'w')

    tkutil.Stoppable.__init__(self, progress_label, br.buttons[4])

  # ---------------------------------------------------------------------------
  #
  def peak_list_box(self, parent):

    pl = tkutil.scrolling_list(parent, 'Possible resonance assignments', 5)
    pl.listbox.bind('<ButtonRelease-1>', self.select_peak_cb)
    pl.listbox.bind('<Double-ButtonRelease-1>', self.goto_peak_cb)
    self.peak_list = pl

    return pl.frame

  # ---------------------------------------------------------------------------
  #
  def select_peak_cb(self, event):

    ed = self.peak_list.event_line_data(event)
    if self.is_assignment(ed):
      epeak, dpeak = ed
      sputil.select_peak(dpeak)

  # ---------------------------------------------------------------------------
  #
  def goto_peak_cb(self, event):

    ed = self.peak_list.event_line_data(event)
    if self.is_assignment(ed):
      epeak, dpeak = ed
      sputil.show_peak(dpeak)
  
  # ---------------------------------------------------------------------------
  # Make assignments associated with selected peak list line.
  # If a peak line is selected the assignment is made and any previous
  # assignments for the expected peak and the data peak are unassigned.
  # If an assignment set header line is selected then the assignments in
  # the set that are not reassignments and that have only one assignment
  # for the expected peak are made.
  #
  def assign_cb(self):

    line_data = self.peak_list.selected_line_data()
    if len(line_data) != 1:
      return

    if self.is_assignment(line_data[0]):
      self.assign_list((line_data[0],))
    elif self.is_assignment_list(line_data[0]):
      assignable = self.unique_assignments(line_data[0])
      self.assign_list(assignable)

  # ---------------------------------------------------------------------------
  # Unassign all assignments associated with selected peak list line.
  # If a peak line is selected the expected peak and data peak assignments
  # (possibly different) are unassigned.  If an assignment set header line
  # is selected then all assignments where the expected peak is assigned to
  # the data peak are unassigned.
  #
  def unassign_cb(self):

    line_data = self.peak_list.selected_line_data()
    if len(line_data) == 0 and self.current_resonance:
      r = self.current_resonance
      ed_list = self.assignments.assignments_for_resonance(r)
      self.unassign_list(ed_list)
    elif len(line_data) == 1:
      if self.is_assignment(line_data[0]):
        self.unassign_list((line_data[0],))
      elif self.is_assignment_list(line_data[0]):
        ed_list = self.exact_assignments(line_data[0])
        self.unassign_list(ed_list)
  
  # ---------------------------------------------------------------------------
  #
  def is_assignment(self, ed):

    return (type(ed) == types.TupleType and
            len(ed) == 2 and
            type(ed[0]) == types.InstanceType and
            type(ed[1]) == types.InstanceType)
  
  # ---------------------------------------------------------------------------
  #
  def is_assignment_list(self, ed_list):

    if type(ed_list) == types.TupleType or type(ed_list) == types.ListType:
      for ed in ed_list:
        if not self.is_assignment(ed):
          return 0
    return 1
  
  # ---------------------------------------------------------------------------
  # Return assignments from list where data peak is unassigned and there is
  # only one assignment in the list for the expected peak.
  #
  def unique_assignments(self, ed_list):

    epeak_count = {}
    for epeak, dpeak in ed_list:
      if not epeak_count.has_key(epeak):
        epeak_count[epeak] = 0
      epeak_count[epeak] = epeak_count[epeak] + 1

    unique = []
    for epeak, dpeak in ed_list:
      if epeak_count[epeak] == 1 and not dpeak.is_assigned:
        unique.append((epeak, dpeak))

    return unique
  
  # ---------------------------------------------------------------------------
  # Return assignments from list where expected peak is assigned to data peak.
  #
  def exact_assignments(self, ed_list):

    exact = []
    for epeak, dpeak in ed_list:
      if self.assignments.has_assignment(epeak, dpeak):
        exact.append((epeak, dpeak))
    return exact
  
  # ---------------------------------------------------------------------------
  #
  def assign_list(self, ed_list):
      
    for epeak, dpeak in ed_list:
      self.assign_peak(epeak, dpeak)

    self.color_atoms()

  # ---------------------------------------------------------------------------
  #
  def assign_peak(self, expected_peak, data_peak):

    self.unassign_peak(expected_peak, data_peak)
    
    # Update assignments object.
    self.assignments.add_peak_assignment(expected_peak, data_peak)

    expected_peak.assign(data_peak)

    # Update spin graph to show newly assigned peak
    self.spin_graph.add_peak(data_peak)
  
  # ---------------------------------------------------------------------------
  #
  def unassign_list(self, ed_list):

    for epeak, dpeak in ed_list:
      self.unassign_peak(epeak, dpeak)

    dpeaks = map(lambda ed: ed[1], ed_list)

    self.color_atoms()

  # ---------------------------------------------------------------------------
  #
  def unassign_peak(self, expected_peak, data_peak):

    # Update assignments object.
    self.assignments.remove_peak_assignment(expected_peak)

    # Unassign peak
    for axis in range(data_peak.spectrum.dimension):
      data_peak.assign(axis, '', '')

    self.spin_graph.remove_peak(data_peak)
    
  # ---------------------------------------------------------------------------
  # Show list of peaks assignable with this group_atom
  #
  def atom_select_cb(self, group_atom):

    eset = self.assignments.experiment_set
    if eset.sequence == None:
      return

    resonance = eset.sequence.group_atom_to_resonance[group_atom]
    self.show_assignments(resonance)

  # ---------------------------------------------------------------------------
  # Show sets of possible assignments for this resonance.
  #
  def show_assignments(self, resonance):

    if self.assignments.is_resonance_assigned(resonance):
      ed_list = (self.assignments.assignments_for_resonance(resonance) +
                 self.assignments.consistent_assignments(resonance))
      ed_list = pyutil.unique_elements(ed_list)
    else:
      ed_list = self.assignments.possible_assignment_list(resonance)

    asets = self.assignment_sets(ed_list)
    asets = pyutil.sort_by_function_value(asets, self.assignment_set_sort_key)
    asets.reverse()

    self.peak_list.clear()
    self.peak_list.heading['text'] = self.assignment_list_heading(resonance)
    for alist, rshifts in asets:
      self.show_assignment_set(alist, rshifts)

    self.current_resonance = resonance
    
  # ---------------------------------------------------------------------------
  #
  def assignment_list_heading(self, resonance):

    heading = 'Resonance %s %s' % (resonance.group_name, resonance.atom_name)
    if self.assignments.is_resonance_assigned(resonance):
      shift = self.assignments.resonance_shift(resonance)
      dev_text = self.deviation_text(resonance, shift)
      heading = heading + ' %6.4g %s' % (shift, dev_text)
    return heading
    
  # ---------------------------------------------------------------------------
  #
  def assignment_set_sort_key(self, aset):

    ed_list, r_shifts = aset

    epeak_table = {}
    for epeak, dpeak in ed_list:
      epeak_table[epeak] = 1

    no_reassignment_epeak_table = {}
    for epeak, dpeak in ed_list:
      if (not dpeak.is_assigned or
          self.assignments.has_assignment(epeak, dpeak)):
        no_reassignment_epeak_table[epeak] = 1

    return (len(epeak_table), len(no_reassignment_epeak_table))
  
  # ---------------------------------------------------------------------------
  #
  def show_assignment_set(self, alist, rshifts):

    rtext = self.resonances_text(rshifts)
    self.peak_list.append(rtext, alist)

    def assignment_sort(ed):
      return (ed[0].experiment.spectrum.name, ed[0].assignment_text())
    ed_list = pyutil.sort_by_function_value(alist, assignment_sort)
                 
    prev_epeak = None
    for epeak, dpeak in ed_list:
      show_assignment = (epeak != prev_epeak)
      atext = self.assignment_text(epeak, dpeak, show_assignment)
      self.peak_list.append(atext, (epeak, dpeak))
      prev_epeak = epeak

    self.peak_list.append('', None)
    
  # ---------------------------------------------------------------------------
  #
  def resonances_text(self, rshifts):

    rtext = ''
    for res, shift in rshifts.items():
      if rtext:
        rtext = rtext + '   '
      rtext = rtext + self.resonance_shift_text(res, shift)
    return rtext
  
  # ---------------------------------------------------------------------------
  #
  def resonance_shift_text(self, resonance, shift):

    gname = resonance.group_name
    aname = resonance.atom_name
    dev = self.deviation_text(resonance, shift)
    return '%s %s %6.4g %s' % (gname, aname, shift, dev)
  
  # ---------------------------------------------------------------------------
  #
  def deviation_text(self, resonance, shift):

    stats = shiftstats.atom_statistics(resonance.symbol, resonance.atom_name)
    if stats == None or stats.shift_deviation == 0:
      return ''

    dev = (shift - stats.average_shift) / stats.shift_deviation

    return '(%+.1f SD)' % dev

  # ---------------------------------------------------------------------------
  #
  def assignment_text(self, epeak, dpeak, show_assignment):
      
    if show_assignment:
      spectrum = epeak.experiment.spectrum.name
      assignment = epeak.assignment_text()
    else:
      spectrum = ''
      assignment = ''
    position = self.position_text(epeak, dpeak)
    assigned = dpeak.assignment
    text = ('%10s %20s %21s %s'
            % (spectrum, assignment, position, assigned))
    return text
      
  # ---------------------------------------------------------------------------
  #
  def position_text(self, epeak, dpeak):

    ptext = ''
    dfreq = dpeak.frequency
    for a in range(len(dfreq)):
      res = epeak.resonances[a]
      rfreq = self.assignments.resonance_shift(res)
      if rfreq == None:
        component = ' %6.4g' % dfreq[a]
      else:
        offset = dfreq[a] - rfreq
        if offset >= 0:
          component = ' +%5.3f' % offset
        else:
          component = ' %6.3f' % offset
      ptext = ptext + component

    return ptext

  # ---------------------------------------------------------------------------
  #
  def assignment_sets(self, assignment_list):

    asets = {}
    for epeak, dpeak in assignment_list:
      rshifts = self.unassigned_shift_table(epeak, dpeak)
      aset = tuple(self.compatible_assignments(assignment_list, rshifts))
      asets[aset] = rshifts
    return asets.items()
  
  # ---------------------------------------------------------------------------
  #
  def unassigned_shift_table(self, epeak, dpeak):

    rshifts = {}
    for axis in range(len(epeak.resonances)):
      r = epeak.resonances[axis]
      if not self.assignments.is_resonance_assigned(r):
        rshifts[r] = dpeak.frequency[axis]
    return rshifts

  # ---------------------------------------------------------------------------
  #
  def compatible_assignments(self, assignment_list, rshifts):

    clist = []
    for epeak, dpeak in assignment_list:
      if self.compatible_assignment(epeak, dpeak, rshifts):
        clist.append((epeak, dpeak))
    return clist
  
  # ---------------------------------------------------------------------------
  #
  def compatible_assignment(self, epeak, dpeak, rshifts):

    for axis in range(len(epeak.resonances)):
      r = epeak.resonances[axis]
      if rshifts.has_key(r):
        efreq = rshifts[r]
        dfreq = dpeak.frequency[axis]
        if abs(dfreq - efreq) > epeak.experiment.tolerances[axis]:
          return 0
      elif not self.assignments.is_resonance_assigned(r):
        return 0
    return 1

  # ---------------------------------------------------------------------------
  #
  def setup_cb(self):

    self.setup.show_window(1)

  # ---------------------------------------------------------------------------
  #
  def update_cb(self):

    if self.assignments:
      eset = self.assignments.experiment_set
      eset.reread_data_peaks()
      self.update_assignments(eset)
      self.update_graph_peaks()

  # ---------------------------------------------------------------------------
  #
  def update_assignments(self, experiment_set):

    self.assignments = possible_assignments(experiment_set, self)

  # ---------------------------------------------------------------------------
  #
  def update_graph_peaks(self):

    self.spin_graph.unplot_all_spectra()
    for e in self.assignments.experiment_set.experiments:
      self.progress_report('Showing peaks for ' + e.spectrum.name)
      self.spin_graph.plot_spectrum(e.spectrum)

    self.color_atoms()

  # ---------------------------------------------------------------------------
  #
  def update_experiment_set(self, experiment_set):

    self.update_assignments(experiment_set)
    self.update_graph_atoms()
    self.update_graph_peaks()
    self.setup.experiments_selected(experiment_set)

    self.progress_report('')

  # ---------------------------------------------------------------------------
  #
  def update_graph_atoms(self):
    
    sgd = self.spin_graph
    if sgd == None:
      sgd = spingraph.spin_graph_dialog()
      sgd.pointer_action.add_atom_select_callback(self.atom_select_cb)
      self.spin_graph = sgd

    sgd.show_window(1)

    eset = self.assignments.experiment_set
    if eset.sequence == None:
      return

    galist = pyutil.attribute_values(eset.resonances, 'group_atom')
    if galist != self.shown_group_atoms:
      self.shown_group_atoms = galist
      self.progress_report('Showing spin graph atoms')
      groups_per_row = eset.sequence.last_residue_number + 1
      sgd.set_graph(galist, groups_per_row, 0, None)
      sgd.graph.update_what_to_show(show_line_label = 0,
                                    lines_to_groups = 0,
                                    shade_by_range = 0)
      sgd.graph.show_atoms(1)
      sgd.graph.full_view('y')

  # ---------------------------------------------------------------------------
  #
  def color_atoms(self):

    sgd = self.spin_graph
    self.progress_report('Coloring atoms')
    for r in self.assignments.assigned_resonances():
      sgd.shade_group_atom(r.group_atom, '')
    for r in self.assignments.unassigned_resonances():
      sgd.shade_group_atom(r.group_atom, 'gray25')
    assignable = self.assignments.assignable_resonances()
    for r in assignable:
      sgd.shade_group_atom(r.group_atom, 'gray50')
    self.progress_report('%d assignable resonances' % len(assignable))

# -----------------------------------------------------------------------------
# Select experiments and tolerances for assigner_dialog.
#
class assigner_setup_dialog(tkutil.Dialog):

  def __init__(self, assigner):

    self.assigner = assigner

    tkutil.Dialog.__init__(self, 'Assignment Graph Setup')

    sc = sputil.spectrum_choices(self.top, 'Use which spectra?', 1, 1)
    sc.frame.pack(side = 'top', anchor = 'w')
    self.choose_spectra = sc

    te = self.tolerance_entry(self.top)
    te.pack(side = 'top', anchor = 'w')
    
    br = tkutil.button_row(self.top,
                           ('Ok', self.ok_cb),
			   ('Apply', self.apply_cb),
                           ('Close', self.close_cb),
                           ('Help', sputil.help_cb('AssignGraph')),
			   )
    br.frame.pack(side = 'top', anchor = 'w')

  # ---------------------------------------------------------------------------
  #
  def tolerance_entry(self, parent):

    t = tkutil.entry_row(parent, 'Tolerances (ppm)',
                         (' 1H', '.02', 5),
                         (' 13C', '.4', 5),
                         (' 15N', '.2', 5))
    self.tolerances = t.variables

    return t.frame

  # ---------------------------------------------------------------------------
  #
  def ok_cb(self):

    self.apply_cb()
    self.close_cb()

  # ---------------------------------------------------------------------------
  #
  def apply_cb(self):

    elist = self.read_experiments()
    eset = experiments.experiment_set(elist)
    self.assigner.update_experiment_set(eset)

  # ---------------------------------------------------------------------------
  #
  def read_experiments(self):

    tolerance_table = self.read_tolerances()
    
    elist = []
    for name in self.choose_spectra.spectrum_names:
      spectrum = sputil.name_to_spectrum(name)
      if spectrum:
        tolerances = pyutil.table_values(tolerance_table, spectrum.nuclei)
        e = experiments.spectrum_experiment(spectrum, tolerances)
        if e == None:
          sputil.message("Don't know peaks to expect for %s\n" % name)
        else:
          elist.append(e)
    return elist

  # ---------------------------------------------------------------------------
  #
  def read_tolerances(self):

    h_tol, c_tol, n_tol = tkutil.float_variable_values(self.tolerances)
    ttable = {'1H':h_tol, '13C':c_tol, '15N':n_tol}
    return ttable

  # ---------------------------------------------------------------------------
  # Update checkbuttons to reflect currently selected spectra
  #
  def experiments_selected(self, eset):

    self.choose_spectra.set_all(0)
    for e in eset.experiments:
      self.choose_spectra.set(e.spectrum, 1)

# -----------------------------------------------------------------------------
# Record peak and resonance assignments.
# This duplicates Sparky's ability to manage peak and resonance assignments.
# It is needed so that sets of conflicting assignments can be remembered.
#
class assignments:

  def __init__(self, experiment_set):

    self.experiment_set = experiment_set
    self.expected_peak_to_data_peak = {}

    #
    # Resonance assignment information is derived from peak assignments.
    # These tables are maintained for efficiency.
    #
    self.resonance_to_assignments = {}
    self.resonance_to_shift = {}
    
  # ---------------------------------------------------------------------------
  #
  def add_peak_assignment(self, expected_peak, data_peak):

    self.remove_peak_assignment(expected_peak)
    self.expected_peak_to_data_peak[expected_peak] = data_peak
    for r in expected_peak.resonances:
      self.add_resonance_assignment(r, expected_peak, data_peak)
    
  # ---------------------------------------------------------------------------
  #
  def remove_peak_assignment(self, expected_peak):

    if self.expected_peak_to_data_peak.has_key(expected_peak):

      data_peak = self.expected_peak_to_data_peak[expected_peak]
      del self.expected_peak_to_data_peak[expected_peak]
      for r in expected_peak.resonances:
        self.remove_resonance_assignment(r, expected_peak)
        
  # ---------------------------------------------------------------------------
  #
  def add_resonance_assignment(self, resonance, expected_peak, data_peak):

    if not self.resonance_to_assignments.has_key(resonance):
      self.resonance_to_assignments[resonance] = {}
    self.resonance_to_assignments[resonance][expected_peak] = data_peak
    self.resonance_to_shift[resonance] = None
        
  # ---------------------------------------------------------------------------
  #
  def remove_resonance_assignment(self, resonance, expected_peak):

    if self.resonance_to_assignments.has_key(resonance):
      e2d = self.resonance_to_assignments[resonance]
      if e2d.has_key(expected_peak):
        del e2d[expected_peak]
        self.resonance_to_shift[resonance] = None
        
  # ---------------------------------------------------------------------------
  #
  def resonance_shift(self, resonance):

    r2s = self.resonance_to_shift
    if not r2s.has_key(resonance) or r2s[resonance] == None:
      r2s[resonance] = self.compute_shift(resonance)
    return r2s[resonance]
        
  # ---------------------------------------------------------------------------
  #
  def compute_shift(self, resonance):

    if not self.resonance_to_assignments.has_key(resonance):
      return None

    total_shift = 0
    shift_count = 0
    for epeak, dpeak in self.resonance_to_assignments[resonance].items():
      for axis in epeak.resonance_axes(resonance):
        total_shift = total_shift + dpeak.frequency[axis]
        shift_count = shift_count + 1

    if shift_count == 0:
      return None

    return total_shift / shift_count

  # ---------------------------------------------------------------------------
  #
  def expected_peak_frequency(self, expected_peak):

    return tuple(map(self.resonance_shift, expected_peak.resonances))
    
  # ---------------------------------------------------------------------------
  #
  def has_assignment(self, expected_peak, data_peak):

    return (self.expected_peak_to_data_peak.has_key(expected_peak) and
            self.expected_peak_to_data_peak[expected_peak] == data_peak)
    
  # ---------------------------------------------------------------------------
  #
  def is_peak_assigned(self, expected_peak):

    return self.expected_peak_to_data_peak.has_key(expected_peak)
    
  # ---------------------------------------------------------------------------
  #
  def is_resonance_assigned(self, resonance):

    r2a = self.resonance_to_assignments
    return (r2a.has_key(resonance) and len(r2a[resonance]) > 0)
    
  # ---------------------------------------------------------------------------
  #
  def are_resonances_assigned(self, resonances):

    r2a = self.resonance_to_assignments
    for r in resonances:
      if not r2a.has_key(r) or len(r2a[r]) == 0:
        return 0
    return 1
    
  # ---------------------------------------------------------------------------
  #
  def any_resonances_assigned(self, resonances):

    r2a = self.resonance_to_assignments
    for r in resonances:
      if r2a.has_key(r) and r2a[r]:
        return 1
    return 0
    
  # ---------------------------------------------------------------------------
  #
  def use_experiment_assignments(self, stoppable):

    for experiment in self.experiment_set.experiments:
      stoppable.progress_report('Calculating possible assignments for ' +
                                experiment.spectrum.name)
      for dpeak in experiment.data_peaks:
        if dpeak.is_assigned:
          epeak = experiment.expected_peak_for_assigned_data_peak(dpeak)
          if epeak:
            self.add_peak_assignment(epeak, dpeak)
    
  # ---------------------------------------------------------------------------
  # Return list of assignments (ie epeak, dpeak pairs) for peaks involving
  # the specified resonance.
  #
  def assignments_for_resonance(self, resonance):

    r2a = self.resonance_to_assignments
    if r2a.has_key(resonance):
      return r2a[resonance].items()
    return []
    
  # ---------------------------------------------------------------------------
  #
  def assigned_resonances(self):

    assigned = []
    for r, e2d in self.resonance_to_assignments.items():
      if e2d:
        assigned.append(r)
    return assigned
    
  # ---------------------------------------------------------------------------
  #
  def unassigned_resonances(self):

    return pyutil.subtract_lists(self.experiment_set.resonances,
                                 self.assigned_resonances())
    
  # ---------------------------------------------------------------------------
  #
  def peak_assignments(self):

    return self.expected_peak_to_data_peak.items()
    
  # ---------------------------------------------------------------------------
  #
  def assigned_peak_count(self):

    return len(self.expected_peak_to_data_peak)
  
# -----------------------------------------------------------------------------
# Find assignments that are consistent with current resonance assignments.
#
class possible_assignments(assignments):

  def __init__(self, experiment_set, stoppable):

    assignments.__init__(self, experiment_set)
    self.use_experiment_assignments(stoppable)

  # ---------------------------------------------------------------------------
  #
  def possible_assignment_list(self, resonance):

    assignments = []
    for epeak in self.experiment_set.expected_peaks(resonance):
      if self.any_resonances_assigned(epeak.resonances):
        freq = self.expected_peak_frequency(epeak)
        dpeaks = epeak.matching_data_peaks(freq)
        for dpeak in dpeaks:
          assignments.append((epeak, dpeak))

    return assignments
  
  # ---------------------------------------------------------------------------
  # Return list of assignments (ie epeak, dpeak pairs) for peaks involving
  # specified resonance for which resonance assignments are already made
  # and there is only one choice within tolerances for assigning the expected
  # peak.
  #
  def consistent_assignments(self, resonance):

    assignments = []
    for epeak in self.experiment_set.expected_peaks(resonance):
      if self.are_resonances_assigned(epeak.resonances):
        freq = self.expected_peak_frequency(epeak)
        dpeaks = epeak.matching_data_peaks(freq)
        if len(dpeaks) == 1:
          assignments.append((epeak, dpeaks[0]))

    return assignments
    
  # ---------------------------------------------------------------------------
  #
  def assignable_resonances(self):

    assigned = {}
    for r in self.assigned_resonances():
      assigned[r] = 1
      
    assignable = {}
    for resonance in assigned.keys():
      for epeak in self.experiment_set.expected_peaks(resonance):
        for r in epeak.resonances:
          if not assigned.has_key(r):
            assignable[r] = 1
    
    return assignable.keys()

# -----------------------------------------------------------------------------
#
def show_experiment_set(eset):
  d = tkutil.the_dialog(assigner_dialog)
  d.show_window(1)
  d.update_experiment_set(eset)

# -----------------------------------------------------------------------------
#
def show_assigner():
  tkutil.the_dialog(assigner_dialog).show_window(1)
