# -----------------------------------------------------------------------------
# Define experiments which combine Sparky spectrum and knowledge about
# expected peaks.
#

import axes
import pyutil
import sequence
import spectrumtype
import sputil

# -----------------------------------------------------------------------------
#
class experiment_set:

  def __init__(self, experiments):

    self.experiments = experiments

    if experiments:
      self.sequence = experiments[0].sequence
    else:
      self.sequence = None
      
    self.resonance_to_expected_peaks = \
      self.resonance_to_expected_peaks_table(experiments)

    self.resonances = self.assignable_resonances(experiments)
  
  # ---------------------------------------------------------------------------
  #
  def resonance_to_expected_peaks_table(self, experiments):

    r2ep = {}

    if self.sequence:
      for r in self.sequence.resonances:
        r2ep[r] = {}
        
    for e in experiments:
      for p in e.expected_peaks:
        for r in p.resonances:
          r2ep[r][p] = 1

    eptable = {}
    for r in r2ep.keys():
      eptable[r] = r2ep[r].keys()

    return eptable
  
  # ---------------------------------------------------------------------------
  #
  def assignable_resonances(self, experiments):

    res = []
    if self.sequence:
      for r in self.sequence.resonances:
        if self.expected_peaks(r):
          res.append(r)

    return res
    
  # ---------------------------------------------------------------------------
  #
  def expected_peaks(self, resonance):

    if self.resonance_to_expected_peaks.has_key(resonance):
      return self.resonance_to_expected_peaks[resonance]
    return []
    
  # ---------------------------------------------------------------------------
  #
  def reread_data_peaks(self):

    for experiment in self.experiments:
      experiment.reread_data_peaks()
      
# -----------------------------------------------------------------------------
# Base class that holds experiment peak list.
# Subclasses for each type of spectrum know how
# to create a list of all expected peaks.
#
class experiment:

  def __init__(self, spectrum, tolerances):

    self.sequence = sequence.molecule_sequence(spectrum.molecule)
    self.spectrum = spectrum
    self.tolerances = tolerances
    self.data_peaks = spectrum.peak_list()
    self.expected_peaks = []
    self.assignment_to_expected_peak = {}
    self.peak_tables = {}

  # ---------------------------------------------------------------------------
  #
  def set_expected_peaks(self, expected_peaks):
    
    for p in expected_peaks:
      p.experiment = self

    a2ep = {}
    for p in expected_peaks:
      a2ep[p.resonances] = p

    self.expected_peaks = expected_peaks
    self.assignment_to_expected_peak = a2ep
    
  # ---------------------------------------------------------------------------
  # Return peaks matching non-None components of frequency to within tolerances
  #
  def nearby_peaks(self, frequency):

    axes = self.search_axes(frequency)
    if not axes:
      return []
    
    f = pyutil.subsequence(frequency, axes)
    if not self.peak_tables.has_key(axes):
      self.peak_tables[axes] = self.peak_table(axes)
    return self.peak_tables[axes].nearby_objects(f)
    
  # ---------------------------------------------------------------------------
  #
  def peak_table(self, axes):

    range = pyutil.subsequence(self.tolerances, axes)
    c = pyutil.cluster(range)
    for p in self.data_peaks:
      f = pyutil.subsequence(p.frequency, axes)
      c.add_point(f, p)
    return c
    
  # ---------------------------------------------------------------------------
  #
  def search_axes(self, frequency):

    axes = []
    for a in range(len(frequency)):
      if frequency[a] != None:
        axes.append(a)
    return tuple(axes)
    
  # ---------------------------------------------------------------------------
  #
  def expected_peak_for_assigned_data_peak(self, data_peak):

    if not data_peak.is_assigned:
      return None

    a = self.sequence.peak_resonances(data_peak)
    if self.assignment_to_expected_peak.has_key(a):
      return self.assignment_to_expected_peak[a]

    return None
    
  # ---------------------------------------------------------------------------
  #
  def reread_data_peaks(self):

    self.data_peaks = self.spectrum.peak_list()
    self.peak_tables = {}
    
# -----------------------------------------------------------------------------
#
class expected_peak:

  def __init__(self, resonances, sign):

    self.experiment = None
    self.sign = sign
    self.resonances = resonances

  # ---------------------------------------------------------------------------
  #
  def assignment_text(self):

    assignment = ''
    last_group = None
    for r in self.resonances:
      group = r.symbol + repr(r.number)
      atom = r.atom_name
      if assignment:
        assignment = assignment + '-'
      if group == last_group:
        assignment = assignment + atom
      else:
        assignment = assignment + group + atom
      last_group = group
    return assignment

  # ---------------------------------------------------------------------------
  #
  def matching_data_peaks(self, freq):
                                            
    dpeaks = []
    for dpeak in self.experiment.nearby_peaks(freq):
      if self.sign > 0:
        correct_sign = (dpeak.data_height > 0)
      elif self.sign < 0:
        correct_sign = (dpeak.data_height < 0)
      else:
        correct_sign = 1
      if correct_sign:
        dpeaks.append(dpeak)
    return dpeaks

  # ---------------------------------------------------------------------------
  #
  def are_resonances_assigned(self, rshifts):

    for r in self.resonances:
      if not rshifts.has_key(r):
        return 0
    return 1

  # ---------------------------------------------------------------------------
  #
  def resonance_axes(self, resonance):

    axes = []
    for axis in range(len(self.resonances)):
      if self.resonances[axis] == resonance:
        axes.append(axis)
    return tuple(axes)

  # ---------------------------------------------------------------------------
  #
  def assign(self, data_peak):

    # Make peak assignment
    rlist = self.resonances
    for axis in range(len(rlist)):
      r = rlist[axis]
      group_name = r.symbol + repr(r.number)
      data_peak.assign(axis, group_name, r.atom_name)

# -----------------------------------------------------------------------------
#
class n15hsqc_experiment(experiment):

  def __init__(self, spectrum, tolerances):

    experiment.__init__(self, spectrum, tolerances)
    epeaks = self.create_expected_peaks(self.spectrum, self.sequence)
    self.set_expected_peaks(epeaks)
    
  # ---------------------------------------------------------------------------
  #
  def create_expected_peaks(self, spectrum, sequence):

    epeaks = []
    for number, symbol in sequence.number_to_symbol.items():
      rH = sequence.resonance(number, symbol, 'H')
      rN = sequence.resonance(number, symbol, 'N')
      if rH and rN:
        resonances = pyutil.permute((rH, rN), axes.hn_axes(spectrum))
        p = expected_peak(resonances, sign = 1)
        epeaks.append(p)
    return epeaks

# -----------------------------------------------------------------------------
#
class hnco_experiment(experiment):

  def __init__(self, spectrum, tolerances):

    experiment.__init__(self, spectrum, tolerances)
    epeaks = self.create_expected_peaks(self.spectrum, self.sequence)
    self.set_expected_peaks(epeaks)

  #----------------------------------------------------------------------------
  #
  def create_expected_peaks(self, spectrum, sequence):

    epeaks = []
    for number, symbol in sequence.number_to_symbol.items():
      add_inter_hnx_peak(spectrum, sequence, number, symbol, 'C', 1, epeaks)
    return epeaks

# -----------------------------------------------------------------------------
#
class hnca_experiment(experiment):

  def __init__(self, spectrum, tolerances):

    experiment.__init__(self, spectrum, tolerances)
    epeaks = self.create_expected_peaks(self.spectrum, self.sequence)
    self.set_expected_peaks(epeaks)

  #----------------------------------------------------------------------------
  #
  def create_expected_peaks(self, spectrum, sequence):

    epeaks = []
    for number, symbol in sequence.number_to_symbol.items():
      add_intra_hnx_peak(spectrum, sequence, number, symbol, 'CA', 1, epeaks)
      add_inter_hnx_peak(spectrum, sequence, number, symbol, 'CA', 1, epeaks)
    return epeaks

# -----------------------------------------------------------------------------
#
class hncoca_experiment(experiment):

  def __init__(self, spectrum, tolerances):

    experiment.__init__(self, spectrum, tolerances)
    epeaks = self.create_expected_peaks(self.spectrum, self.sequence)
    self.set_expected_peaks(epeaks)

  #----------------------------------------------------------------------------
  #
  def create_expected_peaks(self, spectrum, sequence):

    epeaks = []
    for number, symbol in sequence.number_to_symbol.items():
      add_inter_hnx_peak(spectrum, sequence, number, symbol, 'CA', 1, epeaks)
    return epeaks

# -----------------------------------------------------------------------------
#
class cbcaconh_experiment(experiment):

  def __init__(self, spectrum, tolerances):

    experiment.__init__(self, spectrum, tolerances)
    epeaks = self.create_expected_peaks(self.spectrum, self.sequence)
    self.set_expected_peaks(epeaks)
  
  # ---------------------------------------------------------------------------
  #
  def create_expected_peaks(self, spectrum, sequence):

    epeaks = []
    for number, symbol in sequence.number_to_symbol.items():
      add_inter_hnx_peak(spectrum, sequence, number, symbol, 'CA', 1, epeaks)
      add_inter_hnx_peak(spectrum, sequence, number, symbol, 'CB', 1, epeaks)
    return epeaks

# -----------------------------------------------------------------------------
#
class hncacb_experiment(experiment):

  def __init__(self, spectrum, tolerances):

    experiment.__init__(self, spectrum, tolerances)
    epeaks = self.create_expected_peaks(self.spectrum, self.sequence)
    self.set_expected_peaks(epeaks)
  
  # ---------------------------------------------------------------------------
  #
  def create_expected_peaks(self, spectrum, sequence):

    epeaks = []
    for number, symbol in sequence.number_to_symbol.items():
      add_intra_hnx_peak(spectrum, sequence, number, symbol, 'CA', 1, epeaks)
      add_intra_hnx_peak(spectrum, sequence, number, symbol, 'CB', 0, epeaks)
      add_inter_hnx_peak(spectrum, sequence, number, symbol, 'CA', 1, epeaks)
      add_inter_hnx_peak(spectrum, sequence, number, symbol, 'CB', 0, epeaks)
    return epeaks

# -----------------------------------------------------------------------------
#
class hnha_experiment(experiment):

  def __init__(self, spectrum, tolerances):

    experiment.__init__(self, spectrum, tolerances)
    epeaks = self.create_expected_peaks(self.spectrum, self.sequence)
    self.set_expected_peaks(epeaks)

  #----------------------------------------------------------------------------
  #
  def create_expected_peaks(self, spectrum, sequence):

    epeaks = []
    n2s = sequence.number_to_symbol
    for number, symbol in n2s.items():
      if symbol == 'G':
        add_intra_hnx_peak(spectrum, sequence, number, symbol, 'HA2', 1,epeaks)
        add_intra_hnx_peak(spectrum, sequence, number, symbol, 'HA3', 1,epeaks)
      else:
        add_intra_hnx_peak(spectrum, sequence, number, symbol, 'HA', 1, epeaks)
      if n2s.has_key(number-1) and n2s[number-1] == 'G':
        add_inter_hnx_peak(spectrum, sequence, number, symbol, 'HA2', 1,epeaks)
        add_inter_hnx_peak(spectrum, sequence, number, symbol, 'HA3', 1,epeaks)
      else:
        add_inter_hnx_peak(spectrum, sequence, number, symbol, 'HA', 1, epeaks)
    return epeaks

# -----------------------------------------------------------------------------
#
class haconh_experiment(experiment):

  def __init__(self, spectrum, tolerances):

    experiment.__init__(self, spectrum, tolerances)
    epeaks = self.create_expected_peaks(self.spectrum, self.sequence)
    self.set_expected_peaks(epeaks)

  #----------------------------------------------------------------------------
  #
  def create_expected_peaks(self, spectrum, sequence):

    epeaks = []
    n2s = sequence.number_to_symbol
    for number, symbol in n2s.items():
      if n2s.has_key(number-1) and n2s[number-1] == 'G':
        add_inter_hnx_peak(spectrum, sequence, number, symbol, 'HA2', 1,epeaks)
        add_inter_hnx_peak(spectrum, sequence, number, symbol, 'HA3', 1,epeaks)
      else:
        add_inter_hnx_peak(spectrum, sequence, number, symbol, 'HA', 1, epeaks)
    return epeaks
  
# -----------------------------------------------------------------------------
#
def add_intra_hnx_peak(spectrum, sequence, number, symbol,
                       xatom, sign, epeaks):

  rH = sequence.resonance(number, symbol, 'H')
  rN = sequence.resonance(number, symbol, 'N')
  rX = sequence.resonance(number, symbol, xatom)
  if rH and rN and rX:
    resonances = pyutil.permute((rH, rN, rX), axes.amide_axes(spectrum))
    p = expected_peak(resonances, sign)
    epeaks.append(p)
  
# -----------------------------------------------------------------------------
#
def add_inter_hnx_peak(spectrum, sequence, number, symbol,
                       xatom, sign, epeaks):

  prev_number = number - 1
  if not sequence.number_to_symbol.has_key(prev_number):
    return
  
  prev_symbol = sequence.number_to_symbol[prev_number]
  rH = sequence.resonance(number, symbol, 'H')
  rN = sequence.resonance(number, symbol, 'N')
  rX = sequence.resonance(prev_number, prev_symbol, xatom)
  if rH and rN and rX:
    resonances = pyutil.permute((rH, rN, rX), axes.amide_axes(spectrum))
    p = expected_peak(resonances, sign)
    epeaks.append(p)

# -----------------------------------------------------------------------------
#
experiment_class = {
  'n15hsqc':    n15hsqc_experiment,
  'hnco':       hnco_experiment,
  'hnca':       hnca_experiment,
  'hncoca':     hncoca_experiment,
  'hncacb':     hncacb_experiment,
  'cbcaconh':   cbcaconh_experiment,
  'hnha':       hnha_experiment,
  'haconh':     haconh_experiment,
  }

# -----------------------------------------------------------------------------
#
def spectrum_experiment(spectrum, tolerances):

  spectrumtype.guess_spectrum_type(spectrum)
  type = spectrumtype.spectrum_type(spectrum)
  if experiment_class.has_key(type):
    return experiment_class[type](spectrum, tolerances)
  return None
