# -----------------------------------------------------------------------------
# Write a peak list in Mardigras format.
#
import string
import sparky
import sputil
import tkutil

# -----------------------------------------------------------------------------
#
def mardigras_format(peak):
  r1, r2 = proton_resonances(peak)
  if r1 and r1.group.number and r2 and r2.group.number:
    line = "%-4.4s%3d %-4.4s%3d" % (r1.atom.name, r1.group.number,
				    r2.atom.name, r2.group.number)
    if peak.volume:
      line = line + "  %14.4e" % peak.volume
    return line
  return None

# -----------------------------------------------------------------------------
#
def proton_resonances(peak):
  res = peak.resonances()
  if peak.spectrum.dimension == 2:
    return res

  Hres = []
  nuclei = peak.spectrum.nuclei
  for a in range(len(res)):
    if nuclei[a] == '1H':
      Hres.append(res[a])
  return tuple(Hres)

# -----------------------------------------------------------------------------
#
def assignment_order(peak1, peak2):
  r1a, r1b = proton_resonances(peak1)
  r2a, r2b = proton_resonances(peak2)
  return cmp((residue_number_atom(r1a), residue_number_atom(r1b)),
	     (residue_number_atom(r2a), residue_number_atom(r2b)))

# -----------------------------------------------------------------------------
#
def residue_number_atom(r):
  if r and r.group.number:
    return (r.group.number, r.atom.name)
  return None

# -----------------------------------------------------------------------------
#
class peak_list_dialog(tkutil.Dialog):

  # ---------------------------------------------------------------------------
  #
  def __init__(self):

    tkutil.Dialog.__init__(self, 'Peak List')

    self.spectrum_choice = sputil.spectrum_menu(self.top, 'Spectrum: ')
    self.spectrum_choice.frame.pack(side = 'top', anchor = 'w')
    self.spectrum_widget = self.spectrum_choice.frame

    pl = sputil.peak_listbox(self.top)
    pl.frame.pack(side = 'top', fill = 'both', expand = 1)
    pl.listbox.bind('<ButtonRelease-1>', pl.select_peak_cb)
    pl.listbox.bind('<Double-ButtonRelease-1>', pl.goto_peak_cb)
    self.peak_list = pl

    br = tkutil.button_row(self.top,
			   ('Update', self.update_cb),
			   ('Save', self.peak_list.save_cb),
			   ('Close', self.close_cb),
                           )
			   
    br.frame.pack(side = 'top', anchor = 'w')

  # ---------------------------------------------------------------------------
  #
  def update_cb(self):

    spectrum = sputil.name_to_spectrum(self.spectrum_choice.get())

    self.peak_list.clear()

    count = 0
    peaks = spectrum.peak_list()
    peaks.sort(assignment_order)
    for p in peaks:
      line = mardigras_format(p)
      if line:
	self.peak_list.append(line, p)
	count = count + 1

    self.peak_list.heading['text'] = ('%d peaks\n' % count
				      + 'ATOM1   ATOM2   INTENSITY       ')

# -----------------------------------------------------------------------------
#
def show_peak_list():
  tkutil.the_dialog(peak_list_dialog).show_window(1)

# ----------------------------------------------------------------------------
#
short_description = 'make peak list in Mardigras format (ml)'
long_description = (
  'mardigraslist.py - list peaks from a spectrum in Mardigras format\n\n'
  'All assigned peaks are shown.  Unintegrated peaks have intensity\n'
  'left blank.'
)

# -----------------------------------------------------------------------------
#
sparky.add_command('ml', 'Mardigras peak list (ml)',
		   'mardigraslist.show_peak_list()')
