# -----------------------------------------------------------------------------
# Use mirror peaks to help make assignments of 3-D labelled noesy spectra.
# For a pair of close protons H1, H2 you expect two hsqc-noesy peaks.  One
# corresponds to the two protons and the C13 or N15 attached to the first
# proton, and the other is for the two protons and the C13/N15 attached to
# the second proton.  To check a proposed assignment for a peak you can
# look to see if the "mirror peak" (the other peak associated with the pair
# of protons) exists.  This code helps you look for mirror peaks.
# For each possible assignment it lists the signal/noise at the mirror peak
# location and allows you to jump to that place in the spectrum.
#
import string
import Tkinter
import types

import atoms
import axes
import noesy
import pyutil
import sparky
import sputil
import tkutil

class NoHeavyAtomSpectrum(Exception): pass

# -----------------------------------------------------------------------------
#
class mirror_assignment_dialog(tkutil.Dialog, tkutil.Stoppable):

  # ---------------------------------------------------------------------------
  #
  def __init__(self):

    self.last_peak = None
    self.show_selected_cb = None
    self.make_dialog()

  # ---------------------------------------------------------------------------
  #
  def make_dialog(self):

    tkutil.Dialog.__init__(self, 'Mirror Assignment Checker')

    self.top.bind('<Destroy>', self.window_destroyed_cb)

    w = self.make_noesy_table(self.top)
    w.pack(side = 'top', anchor = 'w')

    choices = ('Selected peak assignments',
	       'Peaks with mirror peak',
	       'Peaks without mirror peak',
	       'Unassigned with assigned mirror',
	       )
    self.show_type = tkutil.option_menu(self.top, 'Show ',
                                        choices, choices[0])
    self.show_type.frame.pack(side = 'top', anchor = 'w')

    
    cb = tkutil.checkbuttons(self.top, 'left', 'Show ',
                             'intra', 'sequential', 'long')
    cb.frame.pack(side = 'top', anchor = 'w')
    (self.intra_residue,
     self.sequential,
     self.long_range) = cb.variables

    cb = tkutil.checkbutton(self.top, 'Show alternative assignments?', 0)
    self.alternatives = cb.variable
    cb.button.pack(side = 'top', anchor = 'w')

    e = tkutil.entry_field(self.top, 'C13 / N15 intensity factor: ', '1', 3)
    self.c13_n15_factor = e.variable
    e.frame.pack(side = 'top', anchor = 'w')

    e = tkutil.entry_field(self.top,
                           'Minimum mirror intensity factor: ', '.5', 3)
    self.min_mirror_factor = e.variable
      
    e.frame.pack(side = 'top', anchor = 'w')

    pl = tkutil.scrolling_list(self.top, '', 5)
    pl.frame.pack(side = 'top', fill = 'both', expand = 1)
    pl.listbox.bind('<ButtonRelease-1>', self.select_peak_or_assignment_cb)
    pl.listbox.bind('<Double-ButtonRelease-1>',
                    self.goto_peak_or_assignment_cb)
    pl.listbox.bind('<ButtonPress-2>', tkutil.listbox_select_cb)
    pl.listbox.bind('<ButtonRelease-2>', self.select_mirror_cb)
    pl.listbox.bind('<Double-ButtonRelease-2>', self.goto_mirror_cb)
    self.peak_list = pl

    progress_label = Tkinter.Label(self.top, anchor = 'nw')
    progress_label.pack(side = 'top', anchor = 'w')

    br = tkutil.button_row(self.top,
			   ('Update', self.update_cb),
			   ('Save', self.peak_list.save_cb),
			   ('Stop', self.stop_cb),
			   ('Close', self.close_cb),
                           ('Help', sputil.help_cb('MirrorPeak')),
			   )
    br.frame.pack(side = 'top', anchor = 'w')

    tkutil.Stoppable.__init__(self, progress_label, br.buttons[2])

    self.settings = self.get_settings()

  # ---------------------------------------------------------------------------
  #
  def make_noesy_table(self, parent):

    f = Tkinter.Frame(parent)

    t = Tkinter.Label(f, text = 'noesy spectrum')
    t.grid(row = 1, column = 1)

    t = Tkinter.Label(f, text = 'PPM Tolerance')
    t.grid(row = 0, column = 2, columnspan = 3)
    for a in range(3):
      name = 'w%d' % (a+1)
      h = Tkinter.Label(f, text = name)
      h.grid(row = 1, column = 2 + a)

    self.c13_spect, self.c13_ppm_range = self.noesy_table_row(f, 2, '13C')
    self.n15_spect, self.n15_ppm_range = self.noesy_table_row(f, 3, '15N')

    return f

  # ---------------------------------------------------------------------------
  #
  def noesy_table_row(self, table, row, atom_name):

    t = Tkinter.Label(table, text = atom_name + ' ')
    t.grid(row = row, column = 0)

    spectrum = sputil.spectrum_menu(table, '', allow_no_choice = 1)
    spectrum.frame.grid(row = row, column = 1)

    ppm_range = []
    for a in range(3):
      v = Tkinter.StringVar(table)
      v.set('.01')
      e = Tkinter.Entry(table, textvariable = v, width = 5)
      e.grid(row = row, column = 2 + a)
      ppm_range.append(v)

    s = noesy.noesy_hsqc_spectrum(atom_name)
    if s:
      spectrum.set(s.name)
      heavy_axis = axes.unique_nucleus_axis(s, atom_name)
      ppm_range[heavy_axis].set('.05')

    return spectrum, ppm_range

  # ---------------------------------------------------------------------------
  #
  def window_destroyed_cb(self, event):
    if self.show_selected_cb != None:
      sparky.remove_selection_callback(self.show_selected_cb)
      self.show_selected_cb = None

  # ---------------------------------------------------------------------------
  #
  def update_cb(self):

    self.settings = self.get_settings()

    show_type = self.settings.show_type
    if show_type == 'Peaks with mirror peak':
      self.stoppable_call(self.show_peaks,
			  self.show_peak_with_mirror,
			  'peaks with mirror peak')
    elif show_type == 'Peaks without mirror peak':
      self.stoppable_call(self.show_peaks,
			  self.show_peak_without_mirror,
			  'peaks without mirror peak')
    elif show_type == 'Unassigned with assigned mirror':
      self.stoppable_call(self.show_peaks,
			  self.show_unassigned_with_mirror,
			  'unassigned peaks with assigned mirror peak')
    elif show_type == 'Selected peak assignments':
      if self.show_selected_cb == None:
	self.show_selected_cb = self.show_selected_peak_assignments
	sparky.add_selection_callback(self.show_selected_cb)
      self.show_selected_cb()

    if (self.show_selected_cb != None and
	show_type != 'Selected peak assignments'):
      sparky.remove_selection_callback(self.show_selected_cb)
      self.show_selected_cb = None

  # ---------------------------------------------------------------------------
  #
  def get_settings(self):
    settings = pyutil.generic_class()

    settings.c13_spectrum = sputil.name_to_spectrum(self.c13_spect.get())
    settings.n15_spectrum = sputil.name_to_spectrum(self.n15_spect.get())
    settings.c13_ppm_range = tkutil.float_variable_values(self.c13_ppm_range)
    settings.n15_ppm_range = tkutil.float_variable_values(self.n15_ppm_range)
    settings.show_type = self.show_type.get()
    settings.intra_residue = self.intra_residue.get()
    settings.sequential = self.sequential.get()
    settings.long_range = self.long_range.get()
    settings.alternatives = self.alternatives.get()
    settings.c13_n15_factor = string.atof(self.c13_n15_factor.get())
    try:
      settings.min_mirror_factor = string.atof(self.min_mirror_factor.get())
    except ValueError:
      settings.min_mirror_factor = 0
    return settings

  # ---------------------------------------------------------------------------
  #
  def show_peaks(self, show_peak, description, peaks = None):

    self.peak_list.clear()

    if peaks == None:
      peaks = (self.spectrum_peaks(self.settings.c13_spectrum) +
	       self.spectrum_peaks(self.settings.n15_spectrum))

    count = 0
    self.stoppable_loop('peaks', 50)
    for peak in peaks:
      self.check_for_stop()
      count = count + show_peak(peak)

    heading = '    H        H         Attached     Signal/Noise     '
    self.peak_list.heading['text'] = ('%d %s\n%s' %
				      (count, description, heading))

  # ---------------------------------------------------------------------------
  #
  def spectrum_peaks(self, spectrum):

    peaks = ()
    if spectrum:
      peaks = spectrum.peak_list()
      intra = self.settings.intra_residue
      seq = self.settings.sequential
      long = self.settings.long_range
      if not intra or not seq or not long:
	self.stoppable_loop('peaks filtered', 100)
	peaks = map(self.peak_filter, peaks)
    return pyutil.filter_out_element(peaks, None)

  # ---------------------------------------------------------------------------
  #
  def peak_filter(self, peak):

    self.check_for_stop()
    if not peak.is_assigned:
      return None

    range = sputil.assignment_range(peak)
    if range == 0:
      if self.settings.intra_residue: return peak
    elif range == 1:
      if self.settings.sequential: return peak
    elif self.settings.long_range:
      return peak
    return None

  # ---------------------------------------------------------------------------
  #
  def show_peak_with_mirror(self, peak):

    if not peak.is_assigned:
      return 0

    mirror_assignment, mirror_spectrum, error_message = \
      self.mirror_assignment(peak.resonances(), peak.spectrum)
    if (mirror_assignment == None or
	mirror_spectrum.find_peak(mirror_assignment) == None):
      return 0

    self.show_peak(peak)
    return 1

  # ---------------------------------------------------------------------------
  #
  def show_peak_without_mirror(self, peak):

    if not peak.is_assigned:
      return 0

    mirror_assignment, mirror_spectrum, error_message = \
      self.mirror_assignment(peak.resonances(), peak.spectrum)
    if (mirror_assignment != None and
	mirror_spectrum.find_peak(mirror_assignment) != None):
      return 0

    self.show_peak(peak)
    return 1

  # ---------------------------------------------------------------------------
  #
  def show_unassigned_with_mirror(self, peak):

    if peak.is_assigned:
      return 0

    if not self.is_assignable_with_mirror(peak):
      return 0

    self.show_alternative_assignments(peak)
    return 1

  # ---------------------------------------------------------------------------
  #
  def is_assignable_with_mirror(self, peak):

    alist = self.possible_assignments(peak)
    for a in alist:
      mirror_assignment, mirror_spectrum, error_message = \
        self.mirror_assignment(a, peak.spectrum)
      if (mirror_assignment != None and
	  mirror_spectrum.find_peak(mirror_assignment) != None):
        return 1
    return 0

  # ---------------------------------------------------------------------------
  #
  def mirror_assignment(self, assignment, spectrum = None):

    if not sputil.is_complete_assignment(assignment):
      return None, None, 'partial assignment'

    if spectrum == None:
      spectrum = self.assignment_spectrum(assignment)

    try:
      mirror_assign, mirror_spectrum = \
	mirror_assignment(assignment, spectrum,
			  self.settings.c13_spectrum,
			  self.settings.n15_spectrum)
    except noesy.AtomUnknown, e:
      error_message = 'no Sparky atom %s' % (e.args[0] + e.args[1])
    except noesy.AtomUnassigned, e:
      atom = e.args[0]
      atom_name = atom.group.name + atom.name
      error_message = 'atom %s unassigned' % atom_name
    except noesy.AttachedAtomUnknown, e:
      atom = e.args[0]
      atom_name = atom.group.name + atom.name
      error_message = 'atom attached to %s unknown' % atom_name
    except NoHeavyAtomSpectrum, e:
      atom = e.args[0]
      error_message = 'no spectrum for %s' % atom.name
    else:
      return mirror_assign, mirror_spectrum, None

    return None, None, error_message

  # ---------------------------------------------------------------------------
  #
  def assignment_spectrum(self, assignment):
    c13_spectrum = self.settings.c13_spectrum
    n15_spectrum = self.settings.n15_spectrum
    if c13_spectrum:
      aH, aCN, aH2 = noesy.noesy_hsqc_axes(c13_spectrum)
      r = assignment[aCN]
      if r and atoms.nucleus_type(r.atom) == '13C':
	return c13_spectrum
    if n15_spectrum:
      aH, aCN, aH2 = noesy.noesy_hsqc_axes(n15_spectrum)
      r = assignment[aCN]
      if r and atoms.nucleus_type(r.atom) == '15N':
	return n15_spectrum
    return None

  # ---------------------------------------------------------------------------
  #
  def show_selected_peak_assignments(self):
    self.stoppable_call(self.show_peaks,
			self.show_alternative_assignments,
			'selected peaks', self.selected_peaks())

  # ---------------------------------------------------------------------------
  #
  def selected_peaks(self):
    peaks = []
    if self.settings.c13_spectrum:
      peaks = peaks + self.settings.c13_spectrum.selected_peaks()
    if self.settings.n15_spectrum:
      peaks = peaks + self.settings.n15_spectrum.selected_peaks()
    return peaks

  # ---------------------------------------------------------------------------
  #
  def show_peak(self, peak):

    if self.settings.alternatives:
      self.show_alternative_assignments(peak)
    else:
      self.peak_list.append(self.peak_format(peak), peak)

  # ---------------------------------------------------------------------------
  #
  def show_alternative_assignments(self, peak):
    self.peak_list.append(self.peak_format(peak), peak)
    pa = peak.resonances()
    spectrum = peak.spectrum
    aH, aCN, aH2 = noesy.noesy_hsqc_axes(spectrum)
    alist = self.possible_assignments(peak)
    for a in alist:
      if a != pa:
	omit_labelled_atoms = (pa[aH] == a[aH] and pa[aCN] == a[aCN])
	line = self.assignment_format(a, spectrum, omit_labelled_atoms)
	self.peak_list.append(line, a)
    return 1

  # ---------------------------------------------------------------------------
  #
  def possible_assignments(self, peak):
    if peak.spectrum == self.settings.c13_spectrum:
      ppm_range = self.settings.c13_ppm_range
    elif peak.spectrum == self.settings.n15_spectrum:
      ppm_range = self.settings.n15_ppm_range
    else:
      return None
    alist = noesy.nearby_noesy_hsqc_assignments(peak.spectrum,
						peak.frequency,
						ppm_range)
    if self.min_mirror_factor:
      alist = self.minimum_mirror_intensity_filter(alist, peak)
    return alist

  # ---------------------------------------------------------------------------
  #
  def minimum_mirror_intensity_filter(self, assignments, peak):

    minimum_intensity = self.minimum_mirror_intensity_table(peak)
    alist = []
    for assignment in assignments:
      mirror_assignment, mirror_spectrum, error_message = \
	self.mirror_assignment(assignment, peak.spectrum)
      if mirror_assignment == None:
	alist.append(assignment)
      else:
	intensity = self.signal_to_noise(mirror_assignment, mirror_spectrum)
	if abs(intensity) >= minimum_intensity[mirror_spectrum]:
	  alist.append(assignment)
    return alist

  # ---------------------------------------------------------------------------
  #
  def minimum_mirror_intensity_table(self, peak):

    c13_spectrum = self.settings.c13_spectrum
    n15_spectrum = self.settings.n15_spectrum
    min_factor = self.settings.min_mirror_factor
    c13_n15_factor = self.settings.c13_n15_factor
    peak_min = min_factor * abs(peak.data_height) / peak.spectrum.noise
    minimum_intensity = {}
    if peak.spectrum == c13_spectrum:
      minimum_intensity[c13_spectrum] = peak_min
      minimum_intensity[n15_spectrum] = peak_min / c13_n15_factor
    elif peak.spectrum == n15_spectrum:
      minimum_intensity[c13_spectrum] = peak_min * c13_n15_factor
      minimum_intensity[n15_spectrum] = peak_min
    return minimum_intensity

  # ---------------------------------------------------------------------------
  #
  def signal_to_noise(self, assignment, spectrum):

    peak = spectrum.find_peak(assignment)
    if peak:
      height = peak.data_height
    else:
      pos = sputil.assignment_position(assignment, spectrum)
      height = spectrum.data_height(pos)
    return height / spectrum.noise

  # ---------------------------------------------------------------------------
  #
  def peak_format(self, peak):

    spectrum = peak.spectrum
    assignment = peak.resonances()
    anames = self.atom_names_string(assignment, spectrum)
    peak_signal = peak.data_height / spectrum.noise
    psignal = '%7.1f' % peak_signal
    msignal = self.mirror_signal(assignment, spectrum)

    return ' %s %s %s' % (anames, psignal, msignal)

  # ---------------------------------------------------------------------------
  #
  def assignment_format(self, assignment, spectrum, omit_labelled_atoms):

    anames = self.atom_names_string(assignment, spectrum, omit_labelled_atoms)
    msignal = self.mirror_signal(assignment, spectrum)

    return ' %s %7s %s' % (anames, '', msignal)

  # ---------------------------------------------------------------------------
  #
  def atom_names_string(self, assignment, spectrum, omit_labelled_atoms = 0):

    aH, aCN, aH2 = noesy.noesy_hsqc_axes(spectrum)

    if assignment[aH]:
      group = assignment[aH].group
    else:
      group = None
    h_name = resonance_name(assignment[aH])
    cn_name = resonance_name(assignment[aCN], group)
    h2_name = resonance_name(assignment[aH2], group)
    cn2_name = '?'
    if assignment[aH2]:
      cn2_name = atoms.attached_heavy_atom_name(assignment[aH2].atom)
      if cn2_name == None:
	cn2_name = '?'

    if omit_labelled_atoms:
      return '%-9s %-9s %5s %5s' % ('', h2_name, '', cn2_name)

    return '%-9s %-9s %5s %5s' % (h_name, h2_name, cn_name, cn2_name)

  # ---------------------------------------------------------------------------
  #
  def mirror_signal(self, assignment, spectrum):

    mirror_assignment, mirror_spectrum, error_message = \
      self.mirror_assignment(assignment, spectrum)

    if mirror_assignment:
      signal = self.signal_to_noise(mirror_assignment, mirror_spectrum)
      msignal = '%7.1f' % signal
      if mirror_spectrum.find_peak(mirror_assignment):
	msignal = msignal + ' p'
    else:
      msignal = ''

    return msignal

  # ---------------------------------------------------------------------------
  #
  def select_peak_or_assignment_cb(self, event):

    pora = self.peak_list.event_line_data(event)
    if type(pora) == types.InstanceType:
      sputil.select_peak(pora)
    elif type(pora) == types.TupleType or type(pora) == types.ListType:
      s = self.assignment_spectrum(pora)
      if s:
        sputil.select_peak(s.find_peak(pora))
      
  # ---------------------------------------------------------------------------
  #
  def goto_peak_or_assignment_cb(self, event):

    pora = self.peak_list.event_line_data(event)
    if type(pora) == types.InstanceType:
      sputil.show_peak(pora)
    elif type(pora) == types.TupleType or type(pora) == types.ListType:
      s = self.assignment_spectrum(pora)
      sputil.show_assignment(pora, s)

  # ---------------------------------------------------------------------------
  #
  def select_mirror_cb(self, event):
    apply(sputil.select_assignment, self.clicked_on_mirror(event))

  # ---------------------------------------------------------------------------
  #
  def goto_mirror_cb(self, event):
    apply(sputil.show_assignment, self.clicked_on_mirror(event))

  # ---------------------------------------------------------------------------
  #
  def clicked_on_mirror(self, event):

    pora = self.peak_list.event_line_data(event)
    if type(pora) == types.InstanceType:
      a = pora.resonances()
    elif type(pora) == types.TupleType or type(pora) == types.ListType:
      a = pora
    else:
      a = None
      
    if a:
      s = self.assignment_spectrum(a)
      if s:
	return self.mirror_assignment(a, s)[0:2]
    return None, None

  # ---------------------------------------------------------------------------
  # Zoom a view to position where mirror peak of selected peak should be found
  #
  def show_mirror_peak(self):

    peak = sputil.selected_peak()
    if peak == None:
      return

    if not peak.is_assigned:
      sputil.message('show_mirror_peak(): selected peak is unassigned\n')
      return

    if (self.settings.c13_spectrum == None and
	self.settings.n15_spectrum == None):
      self.show_window(1)	# query for C13 and N15 noesy spectra
      return

    if (peak.spectrum != self.settings.c13_spectrum and
	peak.spectrum != self.settings.n15_spectrum):
      sputil.message('show_mirror_peak(): selected peak is not in spectrum ' +
		     self.spectrum_names() + '\n')
      return

    if peak == self.last_peak:	# repeating command reshows selected peak
      sputil.show_peak(peak)
      self.last_peak = None
      return

    assignment, mirror_spectrum, error_message = \
      self.mirror_assignment(peak.resonances(), peak.spectrum)

    if assignment:
      sputil.show_assignment(assignment, mirror_spectrum)
    else:
      sputil.message('show_mirror_peak(): ' + error_message + '\n')

    self.last_peak = peak

  # ---------------------------------------------------------------------------
  #
  def spectrum_names(self):
    names = ''
    if self.settings.c13_spectrum:
      names = names + ' ' + self.settings.c13_spectrum.name
    if self.settings.n15_spectrum:
      names = names + ' ' + self.settings.n15_spectrum.name
    return names

# ---------------------------------------------------------------------------
#
def resonance_name(resonance, omit_group = None):
  if resonance == None:
    return '?'
  g = resonance.group
  if g == omit_group:
    return resonance.atom.name
  return '%s %s' % (g.name, resonance.atom.name)

# -----------------------------------------------------------------------------
# Return assignment and spectrum for mirror peak.
# Throws AtomUnknown, AtomUnassigned, AttachedAtomUnknown, NoHeavyAtomSpectrum
#
def mirror_assignment(assignment, spectrum, c13_spectrum, n15_spectrum):

  labelled_proton, heavy_atom, unlabelled_proton, other_heavy_atom = \
    noesy.noesy_hsqc_atoms(assignment, spectrum)

  nucleus = atoms.nucleus_type(other_heavy_atom)
  if nucleus == '13C':
    mirror_spectrum = c13_spectrum
  elif nucleus == '15N':
    mirror_spectrum = n15_spectrum
  else:
    raise NoHeavyAtomSpectrum, other_heavy_atom

  a = noesy.noesy_hsqc_assignment(mirror_spectrum, unlabelled_proton,
				  labelled_proton, other_heavy_atom)
  return (a, mirror_spectrum)

# -----------------------------------------------------------------------------
#
def show_mirror_dialog():
  tkutil.the_dialog(mirror_assignment_dialog).show_window(1)

# -----------------------------------------------------------------------------
#
def select_noesy_spectra():
  tkutil.the_dialog(mirror_assignment_dialog).show_window(1)

# -----------------------------------------------------------------------------
#
def show_mirror_peak():
  tkutil.the_dialog(mirror_assignment_dialog).show_mirror_peak()
