# -----------------------------------------------------------------------------
# Show a list of peaks with columns for chemical shifts, volumes, linewidths,
# height, distance bounds, model distances, predicted intensities, ...
#
import os
import Tkinter

import atoms
import corma
import mardigras
import noesy
import pdb
import pyutil
import sputil
import tkutil

# -----------------------------------------------------------------------------
#
class peak_list_dialog(tkutil.Dialog, tkutil.Stoppable):

  # ---------------------------------------------------------------------------
  #
  def __init__(self, spectrum):

    self.spectrum = spectrum
    self.settings = peak_list_settings()

    tkutil.Dialog.__init__(self, spectrum.name + ' peaks')

    pl = sputil.peak_listbox(self.top)
    pl.frame.pack(side = 'top', fill = 'both', expand = 1)
    pl.listbox.bind('<ButtonRelease-1>', pl.select_peak_cb)
    pl.listbox.bind('<Double-ButtonRelease-1>', pl.goto_peak_cb)
    self.peak_list = pl

    progress_label = Tkinter.Label(self.top, anchor = 'nw')
    progress_label.pack(side = 'top', anchor = 'w')

    br = tkutil.button_row(self.top,
			   ('Update', self.update_cb),
			   ('Setup...', self.setup_cb),
			   ('Save...', self.peak_list.save_cb),
			   ('Stop', self.stop_cb),
			   ('Close', self.close_cb),
                           ('Help', sputil.help_cb('PeakListPython')),
			   )
    br.frame.pack(side = 'top', anchor = 'w')

    tkutil.Stoppable.__init__(self, progress_label, br.buttons[3])
  
  # ---------------------------------------------------------------------------
  #
  def update_cb(self):

    self.stoppable_call(self.show_peaks)
  
  # ---------------------------------------------------------------------------
  #
  def setup_cb(self):

    psd = tkutil.the_dialog(peak_list_settings_dialog)
    psd.show_settings(self, self.settings)
    psd.show_window(1)

  # ---------------------------------------------------------------------------
  #
  def new_settings(self, settings):

    self.settings = settings
    self.stoppable_call(self.show_peaks)

  # ---------------------------------------------------------------------------
  #
  def show_peaks(self):
    
    self.progress_report('Getting peaks')
    if self.settings.selected_peaks_only:
      peaks = self.spectrum.selected_peaks()
    else:
      peaks = self.spectrum.peak_list()

    peaks = tkutil.sort_by_key(peaks, peak_sort_key, self)

    if self.settings.pair_crossdiagonal_peaks:
      peaks = pair_crossdiagonal_peaks(peaks, self)

    self.field_initializations(peaks)

    self.peak_list.heading['text'] = ('%d peaks\n' % len(peaks) +
				      self.heading(self.spectrum.dimension))

    self.peak_list.clear()
    self.stoppable_loop('peaks', 50)
    for peak in peaks:
      self.check_for_stop()
      self.peak_list.append(self.peak_line(peak), peak)

  # ---------------------------------------------------------------------------
  #
  def field_initializations(self, peaks):

    for field in self.settings.fields:
      if field.onoff:
	field.initialize(peaks, self)

  # ---------------------------------------------------------------------------
  #
  def peak_line(self, peak):

    line = ''
    for field in self.settings.fields:
      if field.onoff:
	line = line + field.string(peak)
    return line

  # ---------------------------------------------------------------------------
  #
  def heading(self, dim):

    heading = ''
    for field in self.settings.fields:
      if field.onoff:
	heading = heading + field.heading(dim)
    return heading

# -----------------------------------------------------------------------------
#
class peak_list_field:
  def __init__(self):
    self.onoff = self.variable.get()
  def set_dialog_state(self):
    self.variable.set(self.onoff)
  def heading(self, dim):
    if hasattr(self, 'title'):
      return self.pad(self.title(dim), dim)
    return self.pad(self.name, dim)
  def initialize(self, peaks, stoppable):
    pass
  def string(self, peak):
    return self.pad(self.text(peak), peak.spectrum.dimension)
  def pad(self, string, dim):
    size = self.size(dim)
    if size == None:
      return string
    return pyutil.pad_field(string, size)

# -----------------------------------------------------------------------------
# Dialog of possible peak list fields.
#
class peak_list_settings_dialog(tkutil.Dialog):

  # ---------------------------------------------------------------------------
  #
  def __init__(self):

    self.field_classes = []
    self.peak_list_dialog = None

    tkutil.Dialog.__init__(self, 'Peak List Settings')

    fb = Tkinter.Frame(self.top, borderwidth = 3, relief = 'groove')
    fb.pack(side = 'top', fill = 'x')
    
    # -------------------------------------------------------------------------
    #
    class assignment_field(peak_list_field):
      name = 'Assignment'
      def size(self, dim): return 8 * dim
      def text(self, peak): return sputil.assignment_name(peak.resonances())
    self.add_field(fb, assignment_field)

    # -------------------------------------------------------------------------
    #
    class chemical_shift_field(peak_list_field):
      name = 'Chemical Shift'
      def title(self, dim): return 'Shift (ppm)'
      def size(self, dim): return 8 * dim
      def text(self, peak):
	return pyutil.sequence_string(peak.frequency, ' %7.3f')
    self.add_field(fb, chemical_shift_field)

    # -------------------------------------------------------------------------
    #
    class volume_field(peak_list_field):
      name = 'Volume'
      def size(self, dim): return 10
      def text(self, peak):
	if peak.volume: return '%.3g' % peak.volume
	return ''
    self.add_field(fb, volume_field)

    # -------------------------------------------------------------------------
    #
    class data_height_field(peak_list_field):
      name = 'Data Height'
      def title(self, dim): return 'Height'
      def size(self, dim): return 10
      def text(self, peak):
	if peak.data_height == None: return ''
	return '%.3g' % peak.data_height
    self.add_field(fb, data_height_field)

    # -------------------------------------------------------------------------
    #
    class signal_to_noise_field(peak_list_field):
      name = 'Signal / Noise'
      def title(self, dim): return 'S/N'
      def size(self, dim): return 6
      def text(self, peak):
	if peak.data_height == None: return ''
	return '%.1f' % (peak.data_height / peak.spectrum.noise)
    self.add_field(fb, signal_to_noise_field)

    # -------------------------------------------------------------------------
    #
    class fit_height_field(peak_list_field):
      name = 'Fit Height'
      def size(self, dim): return 11
      def text(self, peak):
	if peak.fit_height == None: return ''
	return '%.3g' % peak.fit_height
    self.add_field(fb, fit_height_field)

    # -------------------------------------------------------------------------
    #
    class linewidth_field(peak_list_field):
      name = 'Linewidth'
      def title(self, dim): return 'Linewidth (Hz)'
      def size(self, dim): return 8 * dim
      def text(self, peak):
	if peak.line_width == None: return ''
	linewidth = pyutil.seq_product(peak.line_width,
				       peak.spectrum.hz_per_ppm)
	return pyutil.sequence_string(linewidth, ' %7.2f')
    self.add_field(fb, linewidth_field)

    # -------------------------------------------------------------------------
    #
    class mardigras_bounds_field(peak_list_field):
      name = 'Mardigras Bounds'
      def __init__(self):
	self.path = self.path_widget.get()
	peak_list_field.__init__(self)
      def set_dialog_state(self):
	self.path_widget.set(self.path)
	peak_list_field.set_dialog_state(self)
      def title(self, dim): return '%6s %6s' % ('R-min', 'R-max')
      def size(self, dim): return 14
      def text(self, peak):
	bounds = mardigras.bounds(self.path).distance_bounds(peak)
	if bounds == None: return ''
	return ' %6.2f %6.2f' % bounds
      def initialize(self, peaks, stoppable):
	mardigras.bounds(self.path, stoppable)
    path_widget = tkutil.file_field(self.top, 'File: ', 'mardigras')
    mardigras_bounds_field.path_widget = path_widget
    mardigras_bounds_field.parameter_widget = path_widget.frame
    self.add_field(fb, mardigras_bounds_field)

    # -------------------------------------------------------------------------
    #
    class corma_intensity_field(peak_list_field):
      name = 'Corma Intensities'
      def __init__(self):
	self.path = self.path_widget.get()
	peak_list_field.__init__(self)
      def set_dialog_state(self):
	self.path_widget.set(self.path)
	peak_list_field.set_dialog_state(self)
      def title(self, dim): return 'Corma     '
      def size(self, dim): return 14
      def text(self, peak):
	intensity = corma.intensities(self.path).normalized_intensity(peak)
	if intensity == None: return ''
	if peak.volume:
	  return '%8.3g %4.1f' % (intensity, intensity / peak.volume)
	else:
	  return '%8.3g %4s' % (intensity, '')
      def initialize(self, peaks, stoppable):

	i = corma.intensities(self.path, stoppable)
	for spectrum in sputil.spectra_for_peaks(peaks):
	  i.normalization_factor(spectrum, stoppable)
    path_widget = tkutil.file_field(self.top, 'File: ', 'corma')
    corma_intensity_field.path_widget = path_widget
    corma_intensity_field.parameter_widget = path_widget.frame
    self.add_field(fb, corma_intensity_field)

    # -------------------------------------------------------------------------
    #
    class pdb_distance_field(peak_list_field):
      name = 'PDB Distance'
      def __init__(self):
	self.pdb_path_list = list(self.pdb_files.path_list)
	peak_list_field.__init__(self)
      def set_dialog_state(self):
	self.pdb_files.set_paths(self.pdb_path_list)
	peak_list_field.set_dialog_state(self)
      def title(self, dim):
	h = ''
	for pdb_path in self.pdb_path_list:
	  h = h + ' %6s' % os.path.basename(pdb_path)
	return h
      def size(self, dim): return 7 * len(self.pdb_path_list)
      def text(self, peak):
	f = ''
	if peak.is_assigned:
	  for pdb_path in self.pdb_path_list:
            d = pdb.model(pdb_path).noesy_distance(peak)
	    if d == None:
	      f = f + ' %6s' % ''
	    else:
	      f = f + ' %6.2f' % d
	return f
      def initialize(self, peaks, stoppable):
	for pdb_path in self.pdb_path_list:
	  pdb.model(pdb_path, stoppable)	# read pdb files
    pdb_files = tkutil.file_choices(self.top, 'PDB Files', 'pdb')
    pdb_distance_field.pdb_files = pdb_files
    pdb_distance_field.parameter_widget = pdb_files.frame
    self.add_field(fb, pdb_distance_field)

    # -------------------------------------------------------------------------
    #
    class color_field(peak_list_field):
      name = 'Color'
      def size(self, dim): return 8
      def text(self, peak): return '%8s' % peak.color
    self.add_field(fb, color_field)

    # -------------------------------------------------------------------------
    #
    class note_field(peak_list_field):
      name = 'Note'
      def size(self, dim): return None
      def text(self, peak): return ' ' + peak.note
    self.add_field(fb, note_field)

    # -------------------------------------------------------------------------
    #
    opt = Tkinter.Frame(self.top, borderwidth = 3, relief = 'groove')
    opt.pack(side = 'top', fill = 'x')

    cb = tkutil.checkbutton(opt, 'Selected peaks only?', 0)
    self.selected_only = cb.variable
    cb.button.pack(side = 'top', anchor = 'w')

    cb = tkutil.checkbutton(opt, 'Pair crossdiagonal peaks?', 0)
    self.crosspair = cb.variable
    cb.button.pack(side = 'top', anchor = 'w')

    br = tkutil.button_row(self.top,
			   ('Update', self.update_cb),
			   ('Close', self.close_cb),
			   )
    br.frame.pack(side = 'top', anchor = 'w')

  # ---------------------------------------------------------------------------
  #
  def add_field(self, parent, field_class):

    cb = tkutil.checkbutton(parent, field_class.name, 0)
    field_class.variable = cb.variable
    cb.button.pack(side = 'top', anchor = 'w')

    if hasattr(field_class, 'parameter_widget'):
      cb.map_widget(field_class.parameter_widget)

    self.field_classes.append(field_class)

  # ---------------------------------------------------------------------------
  #
  def show_settings(self, peak_list_dialog, settings):

    self.top.title(peak_list_dialog.spectrum.name + ' peak list settings')
    self.peak_list_dialog = peak_list_dialog
    for f in settings.fields:
      f.set_dialog_state()
    self.selected_only.set(settings.selected_peaks_only)
    self.crosspair.set(settings.pair_crossdiagonal_peaks)

  # ---------------------------------------------------------------------------
  #
  def update_cb(self):

    if self.peak_list_dialog:
      fields = []
      for fc in self.field_classes:
	fields.append(fc())
      settings = peak_list_settings()
      settings.fields = tuple(fields)
      settings.selected_peaks_only = self.selected_only.get()
      settings.pair_crossdiagonal_peaks = self.crosspair.get()

      self.peak_list_dialog.new_settings(settings)

# -----------------------------------------------------------------------------
#
class peak_list_settings:
  def __init__(self):
    self.fields = ()
    self.selected_peaks_only = 0
    self.pair_crossdiagonal_peaks = 0

# -----------------------------------------------------------------------------
# Return a new list of peaks where peaks and crossdiagonal peaks
# are placed consecutively in list.
#
def pair_crossdiagonal_peaks(peaks, stoppable):

  assignment_to_peak = {}
  stoppable.stoppable_loop('finding crossdiagonal peaks', 200)
  for peak in peaks:
    stoppable.check_for_stop()
    r = noesy.proton_resonances(peak)
    assignment_to_peak[(r, peak.spectrum)] = peak

  plist = []
  peak_used = {}
  stoppable.stoppable_loop('sorting crossdiagonal peaks', 200)
  for peak in peaks:
    stoppable.check_for_stop()
    if not peak_used.has_key(peak):
      plist.append(peak)
      peak_used[peak] = 1
      r = noesy.proton_resonances(peak)
      if r:
	tr = (r[1], r[0])
	if assignment_to_peak.has_key((tr, peak.spectrum)):
	  tpeak = assignment_to_peak[(tr, peak.spectrum)]
	  if not peak_used.has_key(tpeak):
	    plist.append(tpeak)
	    peak_used[tpeak] = 1

  return plist

# -----------------------------------------------------------------------------
# Return assignment as ((residue #, residue symbol, atom name), ...)
# for sorting.
#
def peak_sort_key(peak):
  a = []
  for r in peak.resonances():
    if r == None:
      a.append(None)
    else:
      a.append((r.group.number, r.group.symbol, r.atom.name))
  return tuple(a)

# -----------------------------------------------------------------------------
# Show a peak list for the selected spectrum.  If a peak list has already
# been created for the spectrum just make that dialog visible.  Otherwise
# create a new dialog.
#
def show_peak_list():

  spectrum = sputil.selected_spectrum()
  if spectrum:
    if not hasattr(peak_list_dialog, 'spectrum_dialogs'):
      peak_list_dialog.spectrum_dialogs = {}
    dialogs = peak_list_dialog.spectrum_dialogs
    if (dialogs.has_key(spectrum) and
	not dialogs[spectrum].is_window_destroyed()):
      dialogs[spectrum].show_window(1)
    else:
      d = peak_list_dialog(spectrum)
      dialogs[spectrum] = d
      d.show_window(1)
