# ----------------------------------------------------------------------------
# Sparky utility routines and classes
#

import pyutil
import re
import string
import sys
import Tkinter
import tkMessageBox
import tkutil
import types

import sparky

# ----------------------------------------------------------------------------
#
def selected_spectrum():
  v = sparky.selected_view()
  if v:
    return v.spectrum
  return None

# ----------------------------------------------------------------------------
#
def spectrum_list():
  return sparky.current_project().spectrum_list()

# ----------------------------------------------------------------------------
#
def spectrum_names():
  return pyutil.attribute_values(spectrum_list(), 'name')

# ----------------------------------------------------------------------------
#
def name_to_spectrum(name):
  slist = sparky.current_project().spectrum_list()
  for s in slist:
    if s.name == name:
      return s
  return None

# ----------------------------------------------------------------------------
#
def spectra_for_peaks(peaks):
  return pyutil.unique_attribute_values(peaks, 'spectrum')

# ----------------------------------------------------------------------------
#
def selected_condition():
  s = selected_spectrum()
  if s:
    return s.condition
  return None

# ----------------------------------------------------------------------------
#
def selected_condition_name():
  return condition_full_name(selected_condition())

# ----------------------------------------------------------------------------
#
def condition_names():
  return map(condition_full_name, condition_list())

# ----------------------------------------------------------------------------
#
def condition_list():
  proj = sparky.current_project()
  slist = proj.spectrum_list()
  return pyutil.unique_elements(pyutil.attribute_values(slist, 'condition'))

# ----------------------------------------------------------------------------
#
def condition_full_name(c):
  if c and (c.name or c.molecule.name):
    return c.molecule.name + ' / ' + c.name
  return ''

# ----------------------------------------------------------------------------
#
def name_to_condition(name):
  clist = condition_list()
  for c in clist:
    if condition_full_name(c) == name:
      return c
  return None

# ----------------------------------------------------------------------------
# If an assignment along an axis has the same group as the previous axis
# then the group name is not shown.
#
def assignment_name(resonances):
  a = ''
  last_group = None
  for r in resonances:
    if r:
      if r.group == last_group:
	a = a + r.atom.name + '-'
      else:
	a = a + r.name + '-'
      last_group = r.group
    else:
      a = a + '?-'
  return a[:-1]

# -----------------------------------------------------------------------------
# Are all axes unassigned?
#
def is_peak_unassigned(peak):
  for r in peak.resonances():
    if r != None:
      return 0
  return 1

# -----------------------------------------------------------------------------
#
def unassign_spectrum(spectrum):
  for p in spectrum.peak_list():
    rlist = p.resonances()
    for axis in range(spectrum.dimension):
      if rlist[axis]:
        p.assign(axis, '', '')

# -----------------------------------------------------------------------------
#
def assignment_range(peak):
  if not peak.is_assigned:
    return None
  gmin = 1000000
  gmax = -1000000
  for r in peak.resonances():
    if r:
      g = r.group.number
      if g != None:
	if g < gmin: gmin = g
	elif g > gmax: gmax = g
  return gmax - gmin

# -----------------------------------------------------------------------------
#
def molecule_group_atom_list(molecule):
  group_atom_list = []
  for a in molecule.atom_list():
    group_atom_list.append((a.group.name, a.name))
  return group_atom_list

# -----------------------------------------------------------------------------
#
def resonance_group_atom(r):
  return (r.group.name, r.atom.name)

# -----------------------------------------------------------------------------
#
def group_atom_assignment(res):
  return tuple(map(resonance_group_atom, res))

# -----------------------------------------------------------------------------
# Assignment is tuple of (group name, atom name) pairs
#
def group_atom_assignment_name(assignment):
  text = ''
  for group, atom in assignment:
    text = text + group + atom + '-'
  return text[:-1]

# -----------------------------------------------------------------------------
# Assignment is tuple of (group name, atom name) pairs
#
def group_atom_resonances(condition, assignment):
  molecule = condition.molecule
  res = []
  for group_atom in assignment:
    r = group_atom_resonance(group_atom, condition)
    if r == None:
      return None
    res.append(r)
  return res

# -----------------------------------------------------------------------------
#
def group_atom_resonance(group_atom, condition):
  return apply(condition.find_resonance, group_atom)

# -----------------------------------------------------------------------------
#
def number_group_atom(ga):
  return (group_number(ga[0]),) + ga

# -----------------------------------------------------------------------------
#
def group_number(g):
  if g:
    s = re.search('[0-9]+', g)
    if s:
      return string.atoi(s.group())
  return None

# -----------------------------------------------------------------------------
#
def compare_groups(g1, g2):
  return cmp(g1.number, g2.number) or cmp(g1.name, g2.name)

# -----------------------------------------------------------------------------
#
def parse_group_name(g):
  s = re.search('[0-9]+', g)
  if s:
    return g[:s.start()], string.atoi(s.group())
  return g, None

# ---------------------------------------------------------------------------
# Parse assignment of form g1a1-g2a2 to get ((g1, a1), (g2, a2))
# Or g1a1-a2 to get ((g1, a1), (g1, a2))
# A '?' component is translated to ('', '')
#
def parse_assignment(assignment):

  res = string.split(assignment, '-')
  assignment = []
  last_group = None
  for s in res:
    ga = split_group_atom(s)
    if ga == None:
      if last_group:
	ga = (last_group, s)
      else:
	return None
    assignment.append(ga)
    last_group = ga[0]

  return tuple(assignment)

# -----------------------------------------------------------------------------
# Skip to first digit, then skip to first H|C|N to find start of atom name.
#
def split_group_atom(group_atom):

  s = re.search('[0-9]', group_atom)
  if s:
    first_digit = s.start()
    s = re.search('[HCNQM]', group_atom[first_digit:])
    if s:
      HCNQM_offset = s.start()
      d = first_digit + HCNQM_offset
      return (group_atom[:d], group_atom[d:])
  if group_atom == '?':
    return ('', '')
  return None

# -----------------------------------------------------------------------------
# If the peak is off the edge of the spectrum alias it onto the spectrum.
#
def alias_peak(peak):

  dim = peak.spectrum.dimension
  (fmin, fmax) = peak.spectrum.region
  freq = peak.frequency
  alias = dim * [0]
  for a in range(dim):
    if (freq[a] < fmin[a] or freq[a] > fmax[a]):
      position = fmin[a] + (freq[a] - fmin[a]) % (fmax[a] - fmin[a])
      alias[a] = freq[a] - position
  peak.alias = alias

# ----------------------------------------------------------------------------
#
def is_complete_assignment(resonances):
  for r in resonances:
    if r == None:
      return 0
  return 1

# ---------------------------------------------------------------------------
#
def unknown_atom_coordinates(resonances):
  for r in resonances:
    if r and not r.atom.coordinates:
      return 1
  return 0

# -----------------------------------------------------------------------------
#
def selected_peak():
  peaks = sparky.selected_peaks()
  if len(peaks) == 0:
    message('selected_peak(): 0 peaks selected.\n')
  elif len(peaks) > 1:
    message('selected_peak(): 2 or more peaks selected.\n')
  else:
    return peaks[0]
  return None

# -----------------------------------------------------------------------------
#
def show_peak(p):
  if p:
    v = preferred_view(p.spectrum)
    if v:
      v.center = peak_position(p)
      tkutil.raise_named_widget(v.frame)

# -----------------------------------------------------------------------------
#
def select_peak(peak):
  if peak:
    sparky.unselect_all_ornaments()
    peak.selected = 1

# -----------------------------------------------------------------------------
#
def select_assignment(assignment, spectrum):
  if assignment and spectrum:
    select_peak(spectrum.find_peak(assignment))

# -----------------------------------------------------------------------------
#
def show_assignment(assignment, spectrum):
  if assignment and spectrum:
    peak = spectrum.find_peak(assignment)
    if peak:
      show_peak(peak)
    else:
      v = preferred_view(spectrum)
      if v:
        center = assignment_position(assignment, spectrum)
	v.center = center

# -----------------------------------------------------------------------------
#
def peak_position(p):
  pos = []
  for a in range(p.spectrum.dimension):
    pos.append(p.frequency[a] - p.alias[a])
  return tuple(pos)

# -----------------------------------------------------------------------------
# Alias assignment frequency onto spectrum.
#
def assignment_position(assignment, spectrum):

  freq = pyutil.attribute_values(assignment, 'frequency')
  return alias_onto_spectrum(freq, spectrum)

# -----------------------------------------------------------------------------
# Alias frequency onto spectrum.
#
def alias_onto_spectrum(freq, spectrum):
  
  region = spectrum.region
  pos = []
  for a in range(len(freq)):
    min_ppm = region[0][a]
    max_ppm = region[1][a]
    p = min_ppm + (freq[a] - min_ppm) % (max_ppm - min_ppm)
    pos.append(p)
  return tuple(pos)

# -----------------------------------------------------------------------------
# Alias frequency onto spectrum.
#
def alias_axis_onto_spectrum(freq, axis, spectrum):
  
  region = spectrum.region
  min_ppm = region[0][axis]
  max_ppm = region[1][axis]
  return min_ppm + (freq - min_ppm) % (max_ppm - min_ppm)

# -----------------------------------------------------------------------------
#
def closest_peak(freq, peak_list, ranges):

  min_offset = None
  closest = None
  for p in peak_list:
    peak_freq = p.frequency
    offset = 0
    for axis in range(len(freq)):
      offset = offset + abs(peak_freq[axis] - freq[axis]) / ranges[axis]
    if min_offset == None or offset < min_offset:
      min_offset = offset
      closest = p
  return closest

# -----------------------------------------------------------------------------
# 
def resonance_offset_string(frequency, resonances):
  offset = ''
  for a in range(len(frequency)):
    r = resonances[a]
    if r:
      offset = offset + ' %7.4f' % (frequency[a] - r.frequency)
    else:
      offset = offset + ' %7s' % ''
  return offset

# ----------------------------------------------------------------------------
#
def name_to_view(name):
  vlist = sparky.current_project().view_list()
  for v in vlist:
    if v.name == name:
      return v
  return None

# -----------------------------------------------------------------------------
#
def preferred_view(spectrum):
  v = sparky.selected_view()
  if v and v.spectrum == spectrum and v.is_top_level_window:
    return v
  vlist = spectrum.view_list()
  for v in vlist:
    if v.is_shown and v.is_top_level_window:
      return v
  if vlist:
    return vlist[0]
  return None

# -----------------------------------------------------------------------------
#
def hydrogen_atoms(molecule):
  atoms = molecule.atom_list()
  protons = []
  for atom in atoms:
    if atom.name[0] in 'HMQ':
      protons.append(atom)
  return protons

# -----------------------------------------------------------------------------
#
def inter_residue_peaks(peaks, axis1, axis2):

  inter = []
  for peak in peaks:
    r = peak.resonances()
    r1 = r[axis1]
    r2 = r[axis2]
    if r1 and r2:
      n1 = r1.group.number
      n2 = r2.group.number
      if n1 != None and n2 != None and n1 != n2:
	inter.append(peak)
  return inter

# -----------------------------------------------------------------------------
#
def message(m):

  sys.stdout.write(m)
  
# ----------------------------------------------------------------------------
# Print a warning message in a dialog
#
def warn(title, message):

  tkMessageBox.showwarning(title, message, master = tkutil.tk)

# ----------------------------------------------------------------------------
# Create an option menu listing spectra in the current project.
#
def spectrum_menu(parent, label, **kw):

  names = spectrum_names()
  if kw and kw.has_key('allow_no_choice') and kw['allow_no_choice']:
    names = ('',) + names
  selected_spectrum_name = pyutil.attribute(selected_spectrum(), 'name', '')
  return tkutil.option_menu(parent, label, names, selected_spectrum_name)

# ----------------------------------------------------------------------------
# Create an option menu listing views in the current project.
#
def view_menu(parent, label):

  names = pyutil.attribute_values(sparky.current_project().view_list(), 'name')
  selected_view_name = pyutil.attribute(sparky.selected_view(), 'name', '')
  return tkutil.option_menu(parent, label, names, selected_view_name)

# ----------------------------------------------------------------------------
# Used for color coding spectra.  Provides the same colors for spectra to
# all extensions that want to color code (eg spin graphs and strip plots).
#
spectrum_colors = ('green', 'blue', 'red', 'cyan',
                   'magenta', 'yellow', 'orange',
                   'brown', 'gold', 'pink', 'purple')
spectrum_next_color = 0
def spectrum_color(spectrum):

  if not hasattr(spectrum, 'key_color'):
    global spectrum_colors, spectrum_next_color
    spectrum.key_color = spectrum_colors[spectrum_next_color]
    spectrum_next_color = (spectrum_next_color + 1) % len(spectrum_colors)
  return spectrum.key_color

# ----------------------------------------------------------------------------
#
class spectrum_choices:
  
  def __init__(self, parent, label = None, multicolor = 0, onoff = 0):

    f = Tkinter.Frame(parent)
    self.frame = f

    if label:
      h = Tkinter.Label(f, text = label)
      h.pack(side = 'top', anchor = 'w')

    s2b = {}
    names = []
    for spectrum in spectrum_list():
      name = spectrum.name
      cb = tkutil.checkbutton(f, name, onoff)
      s2b[spectrum] = cb
      if multicolor:
        cb.button['selectcolor'] = spectrum_color(spectrum)
      if onoff:
        names.append(name)
      cb.button.pack(side = 'top', anchor = 'w', padx = 20)
      cb.add_callback(pyutil.precompose(choose_spectrum_cb, name, names))

    self.spectrum_to_button = s2b
    self.spectrum_names = names

  # ---------------------------------------------------------------------------
  #
  def set_all(self, onoff):

    for spectrum in self.spectrum_to_button.keys():
      self.set(spectrum, onoff)

  # ---------------------------------------------------------------------------
  #
  def set(self, spectrum, onoff):

    self.spectrum_to_button[spectrum].variable.set(onoff)

# -----------------------------------------------------------------------------
#
def choose_spectrum_cb(name, names, chosen):

  if chosen:
    if not name in names:
      names.append(name)
  else:
    if name in names:
      names.remove(name)

# ----------------------------------------------------------------------------
# Create an option menu listing conditions.
#
def condition_menu(parent, label):

  return tkutil.option_menu(parent, label, condition_names(),
                            selected_condition_name())

# ---------------------------------------------------------------------------
#
class peak_listbox(tkutil.scrolling_list):

  def __init__(self, parent):

    tkutil.scrolling_list.__init__(self, parent, '', 5)

  # ---------------------------------------------------------------------------
  #
  def select_peak_cb(self, event):
    select_peak(self.event_peak(event))

  # ---------------------------------------------------------------------------
  #
  def goto_peak_cb(self, event):
    show_peak(self.event_peak(event))

  # ---------------------------------------------------------------------------
  #
  def event_peak(self, event):
    return self.event_line_data(event)

# ---------------------------------------------------------------------------
#
class assignment_listbox(tkutil.scrolling_list):

  def __init__(self, parent):

    tkutil.scrolling_list.__init__(self, parent, '', 5)

  # ---------------------------------------------------------------------------
  #
  def select_assignment_cb(self, event):
    a = self.event_assignment(event)
    s = selected_spectrum()
    if a and s:
      select_peak(s.find_peak(a))

  # ---------------------------------------------------------------------------
  #
  def goto_assignment_cb(self, event):
    a = self.event_assignment(event)
    show_assignment(a, selected_spectrum())

  # ---------------------------------------------------------------------------
  #
  def event_assignment(self, event):
    return self.event_line_data(event)

# -----------------------------------------------------------------------------
#
def cluster_peaks(peaks, axis1, range1, axis2, range2):

  c = pyutil.cluster((range1, range2))
  for peak in peaks:
    p = (peak.frequency[axis1], peak.frequency[axis2])
    c.add_point(p, peak)

  return c.clusters()

# -----------------------------------------------------------------------------
# Currently doesn't handle function key events.
#
def command_keypress_cb(event):

  sparky.command_characters(event.char)

# -----------------------------------------------------------------------------
# Return a function to be used as a help button callback.
#
def help_cb(extension_topic):

  manual_url = 'extensions.html#' + extension_topic
  return lambda url = manual_url: sparky.show_manual_url(url)
